AceTimer = LibStub("AceTimer-3.0")
local AceComm = LibStub("AceComm-3.0")
local Serializer = LibStub("AceSerializer-3.0")
local Compressor = LibStub("LibCompress")
local Encoder = Compressor:GetAddonEncodeTable()

-- Prefixes
local FULL_MSG_PREFIX = "BL_FULL_MSG"
local SMALL_MSG_PREFIX = "BL_SMALL_MSG"
local ADDON_PREFIX = "BossList"

-- Messages
local REQUEST_UPDATER = "REQ_UPDATER"
local REQUEST_SYNC = "REQ_SYNC"
local AVAILABLE = "AVAILABLE"

local attemptCounter
local requestedUpdater = false 
local playerName = GetUnitName("player", false)
local updaterTimer
local synchronizeTimer
local synchronizeSender
local gottenUpdateMessage = false
local sessionBlacklist = {}

local frame = CreateFrame("Frame")


-- Timer callback methods

local function NoResponses()
  -- No responses means there is nobody online with the addon. Change state to "synchronized"
  AceTimer:CancelTimer(updaterTimer)
  requestedUpdater = false
  BossList.synchronized = true
  --print("Received no responses")
end


-- Checks if the provided player is online. If not, stop waiting for an update and retry synchronization
local function CheckOnline(player)
  GuildRoster()
  
end

-- Send table using prefix
local function SendTable(prefix, table, target) 
  table.revision = BossList.REVISION
  
  local data = Serializer:Serialize(table)
  local compData = Compressor:CompressHuffman(data)
  local encData = Encoder:Encode(compData)
  
  -- If a target is provided, whisper the message. Otherwise broadcast to guild.
  if target == nil then
    AceComm:SendCommMessage(prefix, encData, "GUILD")
  else
    AceComm:SendCommMessage(prefix, encData, "WHISPER", target)
  end
  
  table.revision = nil
end

local function UpdateHandler(prefix, text, distribution, sender)
  if sender == playerName or tContains(sessionBlacklist) then
    return
  end
  
  AceTimer:CancelTimer(synchronizeTimer)
  
  --print("Received update - " .. prefix .. " from " .. sender)
  -- Parse the message
  local compData = Encoder:Decode(text)
  local data, msg = Compressor:Decompress(compData)
  local succes, table = Serializer:Deserialize(data)
  
  if data == nil then
    print("BossList: Failed to decompress - " .. msg)
    print("BossList: Please tell Saberos")
    return
  end
  
  if not succes then
    print("BossList: Failed to deserialize - " .. table)
    print("BossList: Please tell Saberos")
    return
  end
  
  if table.revision == nil or table.revision < BossList.REVISION then
    tinsert(sessionBlacklist, sender)
    print("BossList: Received message from older version from " .. sender .. ". Please tell him/her to update.")
    BossList.Synchronize()
    return
  elseif table.revision > BossList.REVISION then
    if not(gottenUpdateMessage) then
      print("BossList: Received message from newer version. Please update your BossList")
      gottenUpdateMessage = true
    end
    return
  else
    table.revision = nil
  end
  
  -- Update local data
  if prefix == FULL_MSG_PREFIX then
    BossList.MergeData(table)
    BossList.synchronized = true
  elseif prefix == SMALL_MSG_PREFIX then
    BossList.CompareData(table)
  end  
end


local function RequestSync(player)
  requestedUpdater = false
  AceTimer:CancelTimer(updaterTimer)
  AceComm:SendCommMessage(ADDON_PREFIX, REQUEST_SYNC, "WHISPER", player)

  synchronizeTimer = AceTimer:ScheduleRepeatingTimer(CheckOnline, 11)
  synchronizeSender = player
  --print(string.format("Send REQUEST_SYNC to %s", player))
end

local function RequestUpdater()
  requestedUpdater = true
  AceComm:SendCommMessage(ADDON_PREFIX, REQUEST_UPDATER, "GUILD")
  --print("Send REQUEST_UPDATER")
end

local function SendAvailable(player)
  AceComm:SendCommMessage(ADDON_PREFIX, AVAILABLE, "WHISPER", player)
  --print("Send AVAILABLE")
end

local function SyncPlayer(player)
  SendTable(FULL_MSG_PREFIX, BossList_entries, player)
  --print("Send BossList_entries")
end

local function MessageHandler(prefix, text, distribution, sender)  
  -- Ignore messages from the player himself, or players found to running older versions
  if sender == playerName or tContains(sessionBlacklist) then
    return
  end
  
  --print("Received " .. text .. " from " .. sender)
  if text == REQUEST_UPDATER then
    if BossList.synchronized then
      SendAvailable(sender)
    end
  elseif text == REQUEST_SYNC then
    SyncPlayer(sender)
  elseif text == AVAILABLE then
    if requestedUpdater then
      RequestSync(sender)
    end
  end
end

-- Attempt to synchronize 5 times. Multiple attempts are required as on logon chat and guild function calls return
-- false information or simply fail.
local function synchronize()
  if IsInGuild() and GetNumGuildMembers() > 0 then
    
    local _, numOnlineMembers = GetNumGuildMembers()
    if numOnlineMembers > 0  then
      RequestUpdater()
    else  -- If nobody is online, fail immediately
      NoResponses()
    end
    
  end
  
  attemptCounter = attemptCounter + 1
  if attemptCounter >= 5 then
    NoResponses()
  end
end

function BossList.Synchronize()
  attemptCounter = 0
  updaterTimer = AceTimer:ScheduleRepeatingTimer(synchronize, 5)
end

function BossList.SendUpdate(table)
  SendTable(SMALL_MSG_PREFIX, table)
end

local function EventHandler(update)
  if synchronizeTimer ~= nil and AceTimer:TimeLeft(synchronizeTimer) > 0 then
    local numTotalMembers, numOnlineMembers = GetNumGuildMembers()
    local numMembers
    
    if GetGuildRosterShowOffline() then 
      numMembers = numTotalMembers 
    else 
      numMembers = numOnlineMembers
    end
      
    for idx = 1, numMembers do
      local name, _, _, _, _, _, _, _, online, _, _, _, _, _ = GetGuildRosterInfo(idx)
      if online and name == synchronizeSender then
        --print(synchronizeSender .. " is still online")
        return
      end
    end
    
    --print(synchronizeSender .. " disconnected")
    BossList.Synchronize()  -- Player disconnected during transaction, retry synchronization
  end
end


AceComm:RegisterComm(FULL_MSG_PREFIX, UpdateHandler)
AceComm:RegisterComm(SMALL_MSG_PREFIX, UpdateHandler)
AceComm:RegisterComm(ADDON_PREFIX, MessageHandler)

frame:RegisterEvent("GUILD_ROSTER_UPDATE")
frame:SetScript("OnEvent", EventHandler)